"""
DIANNA: Deep Insight And Neural Network Analysis.

Modern scientific challenges are often tackled with (Deep) Neural Networks (DNN).
Despite their high predictive accuracy, DNNs lack inherent explainability. Many DNN
users, especially scientists, do not harvest DNNs power because of lack of trust and
understanding of their working.

Meanwhile, the eXplainable AI (XAI) methods offer some post-hoc interpretability and
insight into the DNN reasoning. This is done by quantifying the relevance of individual
features (image pixels, words in text, etc.) with respect to the prediction. These
"relevance heatmaps" indicate how the network has reached its decision directly in the
input modality (images, text, speech etc.) of the data.

There are many Open Source Software (OSS) implementations of these methods, alas,
supporting a single DNN format and the libraries are known mostly by the AI experts.
The DIANNA library supports the best XAI methods in the context of scientific usage
providing their OSS implementation based on the ONNX standard and demonstrations on
benchmark datasets. Representing visually the captured knowledge by the AI system can
become a source of (scientific) insights.

See https://github.com/dianna-ai/dianna
"""
import importlib
import logging
from . import utils


logging.getLogger(__name__).addHandler(logging.NullHandler())

__author__ = "DIANNA Team"
__email__ = "dianna-ai@esciencecenter.nl"
__version__ = "0.4.2"


def explain_image(model_or_function, input_data, method, labels=(1,), **kwargs):
    """
    Explain an image (input_data) given a model and a chosen method.

    Args:
        model_or_function (callable or str): The function that runs the model to be explained _or_
                                             the path to a ONNX model on disk.
        input_data (np.ndarray): Image data to be explained
        method (string): One of the supported methods: RISE, LIME or KernelSHAP
        labels (tuple): Labels to be explained

    Returns:
        One heatmap (2D array) per class.

    """
    if method == "KernelSHAP":
        # To avoid Access Violation on Windows with SHAP:
        from onnx_tf.backend import prepare  # pylint: disable=import-outside-toplevel,unused-import
    explainer = _get_explainer(method, kwargs)
    explain_image_kwargs = utils.get_kwargs_applicable_to_function(explainer.explain_image, kwargs)
    return explainer.explain_image(model_or_function, input_data, labels, **explain_image_kwargs)


def explain_text(model_or_function, input_data, method, labels=(1,), **kwargs):
    """
    Explain text (input_data) given a model and a chosen method.

    Args:
        model_or_function (callable or str): The function that runs the model to be explained _or_
                                             the path to a ONNX model on disk.
        input_data (string): Text to be explained
        method (string): One of the supported methods: RISE or LIME
        labels (tuple): Labels to be explained

    Returns:
        List of (word, index of word in raw text, importance for target class) tuples.

    """
    explainer = _get_explainer(method, kwargs)
    explain_text_kwargs = utils.get_kwargs_applicable_to_function(explainer.explain_text, kwargs)
    return explainer.explain_text(model_or_function, input_data, labels, **explain_text_kwargs)


def _get_explainer(method, kwargs):
    method_submodule = importlib.import_module(f'dianna.methods.{method.lower()}')
    method_class = getattr(method_submodule, method)
    method_kwargs = utils.get_kwargs_applicable_to_function(method_class.__init__, kwargs)
    return method_class(**method_kwargs)



import inspect


def get_function(model_or_function, preprocess_function=None):
    """Converts input to callable function.

    Any keyword arguments are given to the ModelRunner class if the input is a model path.

    Args:
        model_or_function: Can be either model path or function.
            If input is a function, the function is returned unchanged.
        preprocess_function: function to be run to preprocess the data
    """
    from dianna.utils.onnx_runner import SimpleModelRunner  # pylint: disable=import-outside-toplevel
    if isinstance(model_or_function, str):
        runner = SimpleModelRunner(model_or_function, preprocess_function=preprocess_function)
    elif callable(model_or_function):
        if preprocess_function is None:
            runner = model_or_function
        else:
            def runner(input_data):
                return model_or_function(preprocess_function(input_data))
    else:
        raise TypeError("model_or_function argument must be string (path to model) or function")
    return runner


def get_kwargs_applicable_to_function(function, kwargs):
    """Returns a subset of `kwargs` of only arguments and keyword arguments of `function`.

    Note that if `function` has a `**kwargs`
    argument, this function should not be necessary (provided the function
    handles `**kwargs` robustly).
    """
    return {key: value for key, value in kwargs.items()
            if key in inspect.getfullargspec(function).args}


def to_xarray(data, axis_labels, required_labels=None):
    """Converts numpy data and axes labels to an xarray object."""
    if isinstance(axis_labels, dict):
        # key = axis index, value = label
        # not all axes have to be present in the input, but we need to provide
        # a name for each axis
        # first ensure negative indices are converted to positive ones
        indices = list(axis_labels.keys())
        for index in indices:
            if index < 0:
                axis_labels[data.ndim + index] = axis_labels.pop(index)
        labels = [axis_labels[index] if index in axis_labels else f'dim_{index}' for index in range(data.ndim)]
    else:
        labels = list(axis_labels)

    # check if the required labels are present
    if required_labels is not None:
        for label in required_labels:
            assert label in labels, f'Required label missing: {label}'

    # import here because it's slow
    import xarray as xr  # pylint: disable=import-outside-toplevel

    return xr.DataArray(data, dims=labels)


def move_axis(data, label, new_position):
    """Moves a named axis to a new position in an xarray DataArray object.

    Args:
        data (DataArray): Object with named axes
        label (str): Name of the axis to move
        new_position (int): Numerical new position of the axis.
                            Negative indices are accepted.

    Returns:
        data with axis in new position
    """
    # find current position of axis
    try:
        pos = data.dims.index(label)
    except ValueError as e:
        raise ValueError(f"Axis name {label} does not exist in input data") from e

    # create list of labels with new ordering
    axis_labels = list(data.dims)
    # the new position will be _before_ the given index, so will fail with a negative index
    # convert to a positive index in that case
    if new_position < 0:
        new_position += len(axis_labels)
    axis_labels.insert(new_position, axis_labels.pop(pos))
    # do the move
    return data.transpose(*axis_labels)


def onnx_model_node_loader(model_path):
    """Onnx model and node labels loader.

    Load onnx model and return the label of its output node and the data type of input node.

    Args:
        model_path (str): The path to a ONNX model on disk.

    Returns:
        loaded onnx model and the label of output node.
    """
    # these imports are done in the function because they are slow
    import onnx  # pylint: disable=import-outside-toplevel
    from onnx_tf.backend import prepare  # pylint: disable=import-outside-toplevel
    onnx_model = onnx.load(model_path)  # load onnx model
    tf_model_rep = prepare(onnx_model, gen_tensor_dict=True)
    label_input_node = tf_model_rep.inputs[0]
    label_output_node = tf_model_rep.outputs[0]
    dtype_input_node = tf_model_rep.tensor_dict[f'{label_input_node}'].dtype

    return onnx_model, dtype_input_node, label_output_node
